// Copyright (c) 2015 HigPen Works (Kenichiro Matsuura / Yuki Tsukasa). All rights reserved.
#include "Main.h"
#include "Stage.h"

bool UseScore;
int Score;

bool ResetStage;
int StageIndex=0;

CMoverList StageList(sizeof(CStage));
CMoverList CameraList(sizeof(CCamera));
CMoverList EffectList(sizeof(CEffect));
CMoverList ItemList(sizeof(CItem));
CMoverList BulletList(sizeof(CBullet));
CMoverList WeaponList(sizeof(CWeapon));
CMoverList EnemyList(sizeof(CEnemy));
CMoverList PlayerList(sizeof(CPlayer));
CMoverList BlockList(sizeof(CBlock));
CMoverList TextList(sizeof(CText));

extern CMoverList UIList;
class CUI : public CMover {
public:
	static CMoverList& MoverList() { return UIList; }
	CVector BasePosition;
	void Reset() {
		const wchar_t* texture[]={L"Left.png", L"Center.png", L"Right.png"};
		Texture=GetTexture(texture[Type]);
		Scale=CVector(0.1f, 0.1f);
		BasePosition=Position;
	}
	void Move() {
		Position=Game->Camera->Position+BasePosition;
		if (!OldTouch[0] && Touch[0]) {
			if (Length(Game->Camera->Position+Touch[0].Position-Position)<Scale.X*1.5f) {
				ResetStage=true;
				switch (Type) {
					case 0: StageIndex--; break;
					case 2: StageIndex++; break;
				}
			}
		}
	}
};
CMoverList UIList(sizeof(CUI));

class CMyGame : public CGame {
public:
	CMyGame() {
		StageIndex=Config.GetInt(L"StageIndex");
	}
	void Reset() {
		Config.Set(L"StageIndex", StageIndex);
		ResetStage=true;
		Scale=CVector(1, 1.5f, 0);

		CameraList.Kill();
		StageList.Kill();
		PlayerList.Kill();
		EnemyList.Kill();
		EffectList.Kill();
		BulletList.Kill();
		WeaponList.Kill();
		BlockList.Kill();
		ItemList.Kill();
		TextList.Kill();
		UIList.Kill();

		UseScore=false;
		Score=0;
		Camera=New<CCamera>();
		StageFunc[StageIndex]();

		New<CUI>(0, CVector(-0.8f, -1.4f));
		New<CUI>(1, CVector(-0.4, -1.4f));
		New<CUI>(2, CVector(0, -1.4f));
	}
	void Move() {
		for (int i=0, s=StageIndex, n=max(1, (int)log10(s)+1); i<n; i++, s/=10) {
			New<CText>(L'0'+s%10, Camera->Position+CVector(0.9f-0.2f*i, -1.4f, 0));
		}
		if (ResetStage && Touch[0]) return;

		CameraList.Move();
		StageList.Move();
		PlayerList.Move();
		EnemyList.Move();
		EffectList.Move();
		BulletList.Move();
		WeaponList.Move();
		ItemList.Move();
		BlockList.Move();

		if (UseScore) {
			for (int i=0, s=Score, n=max(1, (int)log10(s)+1); i<n; i++, s/=10) {
				New<CText>(L'0'+s%10, Camera->Position+CVector(0.9f-0.2f*i, 1.3f, 0));
			}
		}
		TextList.Move();
		
		ResetStage=false;
		UIList.Move();
		if (ResetStage) StageIndex=(StageIndex+StageCount)%StageCount, Reset();
	}
	void Draw() {
		Clear(CColor(1, 1, 1, 1));
		StageList.Draw();
		EffectList.Draw();
		BlockList.Draw();
		EnemyList.Draw();
		WeaponList.Draw();
		PlayerList.Draw();
		BulletList.Draw();
		ItemList.Draw();
		TextList.Draw();
		UIList.Draw();
	}
};

CGame* Main() {
	return new CMyGame();
}
