// Copyright (c) 2015 HigPen Works (Kenichiro Matsuura / Yuki Tsukasa). All rights reserved.
#include "Main.h"

//---------------------------------------------------------
class CPlayerFall : public CPlayer {
public:
	void Move() {
		Velocity.Y-=0.001f;
		Position+=Velocity;
	}
};
class CStageFall : public CStage {
public:
	void Reset() {
		New<CPlayerFall>(CVector(0, 1));
	}
};
CStage* StageFall() { return New<CStageFall>(); }

//---------------------------------------------------------
class CPlayerFloor : public CPlayer {
public:
	void Move() {
		Velocity.Y-=0.001f;
		Position+=Velocity;
		CHitBox(this, BlockList, 0.05f);
	}
};
class CStageFloor : public CStage {
public:
	void Reset() {
		New<CPlayerFloor>(CVector(0, 1));
		New<CBlock>(CVector(0, -1));
	}
};
CStage* StageFloor() { return New<CStageFloor>(); }

//---------------------------------------------------------
class CPlayerJump : public CPlayer {
public:
	void Move() {
		Velocity.Y-=0.001f;
		Position+=Velocity;
		if (CHitBox(this, BlockList, 0.05f).Down && Touch[0]) Velocity.Y=0.06f;
	}
};
class CStageJump : public CStage {
public:
	void Reset() {
		New<CPlayerJump>(CVector(0, 1));
		New<CBlock>(CVector(0, -1));
	}
};
CStage* StageJump() { return New<CStageJump>(); }

//---------------------------------------------------------
class CPlayerJumpHeight : public CPlayer {
public:
	void Move() {
		Velocity.Y-=0.001f;
		Position+=Velocity;
		if (CHitBox(this, BlockList, 0.05f).Down && !Touch[0] && OldTouch[0]) Velocity.Y=min(OldTouch[0].Time, 60)*0.001f;
	}
};
class CStageJumpHeight : public CStage {
public:
	void Reset() {
		New<CPlayerJumpHeight>(CVector(0, 1));
		New<CBlock>(CVector(0, -1));
	}
};
CStage* StageJumpHeight() { return New<CStageJumpHeight>(); }

//---------------------------------------------------------
class CPlayerJumpMultiple : public CPlayer {
public:
	void Move() {
		Velocity.Y-=0.001f;
		Position+=Velocity;
		if (CHitBox(this, BlockList, 0.05f).Down) State=2;
		if (Touch[0] && !OldTouch[0] && State>0) {
			Velocity.Y=0.04f;
			State--;
		}
	}
};
class CStageJumpMultiple : public CStage {
public:
	void Reset() {
		New<CPlayerJumpMultiple>(CVector(0, 1));
		New<CBlock>(CVector(0, -1));
	}
};
CStage* StageJumpMultiple() { return New<CStageJumpMultiple>(); }

//---------------------------------------------------------
class CPlayerJumpParabola : public CPlayer {
public:
	void Move() {
		Velocity.X=0.005f;
		Velocity.Y-=0.001f;
		Position+=Velocity;
		if (CHitBox(this, BlockList, 0.05f).Down && Touch[0]) Velocity.Y=0.06f;
	}
};
class CStageJumpParabola : public CStage {
public:
	void Reset() {
		New<CPlayerJumpParabola>(CVector(-1, -0.8f));
		for (int i=-5; i<=5; i++) New<CBlock>(CVector(i*0.2f, -1));
	}
};
CStage* StageJumpParabola() { return New<CStageJumpParabola>(); }

//---------------------------------------------------------
class CPlayerFloorDrop : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Velocity.X=0.02f; else Velocity.X=0;
		Velocity.Y-=0.001f;
		Position+=Velocity;
		BlockList.Apply([&](CMover* m) {
			if (CHitBox(this, m, 0.05f).Down) m->State=1;
		});
	}
};
class CBlockFloorDrop : public CBlock {
public:
	void Move() {
		if (State==1) {
			Time++;
			if (Time>10) State=2;
		}
		if (State==2) {
			Velocity.Y-=0.001f;
			Position+=Velocity;
		}
	}
};
class CStageFloorDrop : public CStage {
public:
	void Reset() {
		New<CPlayerFloorDrop>(CVector(-1.01f, 0.2f));
		for (int i=-4; i<=4; i++) New<CBlockFloorDrop>(CVector(i*0.2f, 0));
		New<CBlock>(CVector(1, 0))->Texture=GetTexture(L"Block2.png");
		New<CBlock>(CVector(-1, 0))->Texture=GetTexture(L"Block2.png");
	}
};
CStage* StageFloorDrop() { return New<CStageFloorDrop>(); }

//---------------------------------------------------------
class CPlayerFloorSlide : public CPlayer {
public:
	void Move() {
		Velocity.X=0;
		BlockList.Apply([&](CMover* m) {
			if (CHitBox(this, m, 0.05f).Down) Velocity.X=m->Velocity.X;
		});
		if (Touch[0]) Velocity.X+=0.02f;
		Velocity.Y-=0.001f;
		Position+=Velocity;
	}
};
class CBlockFloorSlide : public CBlock {
public:
	void Reset() {
		CBlock::Reset();
		Velocity.X=0.01f;
		Scale.X=0.3f;
	}
	void Move() {
		Position+=Velocity;
		if (abs(Position.X)>0.5f) Velocity.X*=-1;
	}
};
class CStageFloorSlide : public CStage {
public:
	void Reset() {
		New<CPlayerFloorSlide>(CVector(-0.9f, 0.2f));
		New<CBlockFloorSlide>(CVector(0, 0));
		New<CBlock>(CVector(-0.9f, 0))->Texture=GetTexture(L"Block2.png");
		New<CBlock>(CVector(0.9f, 0))->Texture=GetTexture(L"Block2.png");
	}
};
CStage* StageFloorSlide() { return New<CStageFloorSlide>(); }

//---------------------------------------------------------
class CPlayerStepOn : public CPlayer {
public:
	void Move() {
		Velocity.X=0.005f;
		Velocity.Y-=0.001f;
		Position+=Velocity;
		if (CHitBox(this, BlockList, 0.05f).Down && Touch[0]) Velocity.Y=0.04f;
		EnemyList.Apply([&](CMover* m) {
			if (Velocity.Y<0 && CHitBox(this, m, 0.05f).Down) Velocity.Y=0.03f, m->State=1;
		});
	}
};
class CEnemyStepOn : public CEnemy {
public:
	void Move() {
		if (State==1) {
			Rotation*=RotationZ(0.02f);
			Color.A*=0.9f;
			if (Color.A<0.1f) Life=0;
		}
	}
};

class CStageStepOn : public CStage {
public:
	void Reset() {
		New<CPlayerStepOn>(CVector(-1, 0));
		for (int i=-2; i<=2; i++) New<CEnemyStepOn>(CVector(i*0.3f, 0));
		for (int i=-5; i<=5; i++) New<CBlock>(CVector(i*0.2f, -0.2f));
	}
};
CStage* StageStepOn() { return New<CStageStepOn>(); }

//---------------------------------------------------------
class CPlayerTrampoline : public CPlayer {
public:
	void Move() {
		Velocity.X=0.005f;
		Velocity.Y-=0.001f;
		Position+=Velocity;
		BlockList.Apply([&](CMover* m) {
			if (CHitBox(this, m, 0.05f).Down && Touch[0]) {
				if (m->Type==1) Velocity.Y=0.04f; else Velocity.Y=0.02f;
			}
		});
	}
};
class CBlockTrampoline : public CBlock {
public:
	void Reset() {
		CBlock::Reset();
		if (Type==0) Texture=GetTexture(L"Block2.png");
		if (Type==1) Texture=GetTexture(L"Block.png"), Scale.Y=0.05f, Position.Y+=Scale.X-Scale.Y;
	}
};

class CStageTrampoline : public CStage {
public:
	void Reset() {
		New<CPlayerTrampoline>(CVector(-1, 0.2f));
		for (int i=-5; i<=5; i++) New<CBlockTrampoline>(abs(i)<=1?1:0, CVector(i*0.2f, 0));
	}
};
CStage* StageTrampoline() { return New<CStageTrampoline>(); }

//---------------------------------------------------------
class CPlayerItemPick : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
		ItemList.Apply([&](CMover* m) {
			if (Length(Position-m->Position)<0.2f) m->Life=0, Score++;
		});
	}
};

class CStageItemPick : public CStage {
public:
	void Reset() {
		UseScore=true;
		New<CPlayerItemPick>();
		for (int i=0; i<10; i++) New<CItem>(CVector(Random(-0.9f, 0.9f), Random(-1.4f, 1.4f)));
	}
};
CStage* StageItemPick() { return New<CStageItemPick>(); }

//---------------------------------------------------------
class CPlayerItemStrongWeak : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
		ItemList.Apply([&](CMover* m) {
			if (Length(Position-m->Position)<0.2f) {
				m->Life=0;
				if (m->Type==0) Scale*=1.5f; else Scale/=1.5f;
			}
		});
	}
};
class CItemStrongWeak : public CItem {
public:
	void Reset() {
		CItem::Reset();
		Scale=Type==0?0.1f:0.05f;
	}
};
class CStageItemStrongWeak : public CStage {
public:
	void Reset() {
		New<CPlayerItemStrongWeak>();
		for (int i=0; i<10; i++) New<CItemStrongWeak>(i%2, CVector(Random(-0.9f, 0.9f), Random(-1.4f, 1.4f)));
	}
};
CStage* StageItemStrongWeak() { return New<CStageItemStrongWeak>(); }

//---------------------------------------------------------
class CItemThrow : public CItem {
public:
	void Move() {
		switch (State) {
		case 0:
			Position=Owner->Position+CVector(0, 0.2f);
			if (Touch[0]) State=1;
			break;
		case 1:
			Position.Y+=0.03f;
			if (Position.Y>2) Life=0;
			break;
		}
	}
};
class CPlayerItemThrow : public CPlayer {
public:
	void Reset() {
		CPlayer::Reset();
		New<CItemThrow>()->Owner=this;
		Velocity.X=0.02f;
	}
	void Move() {
		Position+=Velocity;
		if (abs(Position.X)>1) Velocity.X*=-1;
	}
};
class CStageItemThrow : public CStage {
public:
	void Reset() {
		New<CPlayerItemThrow>(CVector(0, -0.5f));
	}
};
CStage* StageItemThrow() { return New<CStageItemThrow>(); }

//---------------------------------------------------------
class CItemSling : public CItem {
public:
	void Move() {
		CVector center(0, 0), v=center-Position;
		switch (State) {
		case 0:
			Owner->Scale.Y=0.1f;
			if (Touch[0]) State=1;
			break;
		case 1:
			Position=Touch[0].Position;
			Owner->Scale.Y=(Length(v)-0.1f)/2;
			if (!Touch[0]) Velocity=v*0.05f, State=2;
			break;
		case 2:
			Position+=Velocity;
			Owner->Scale.Y*=0.9f;
			break;
		}
		Owner->Rotation=CQuaternion(-AXIS_Y, v, AXIS_Z);
		Owner->Position=center+AXIS_Y*Owner->Scale.Y*Owner->Rotation;
	}
};
class CPlayerSling : public CPlayer {
public:
	void Reset() {
		CPlayer::Reset();
		New<CItemSling>(CVector(0, 0.3f))->Owner=this;
	}
};
class CStageSling : public CStage {
public:
	void Reset() {
		New<CPlayerSling>();
	}
};
CStage* StageSling() { return New<CStageSling>(); }

//---------------------------------------------------------
class CPlayerReboundScreen : public CPlayer {
public:
	void Move() {
		Position+=Velocity;
		if (abs(Position.X)>1) Velocity.X*=-1;
		if (abs(Position.Y)>1.5f) Velocity.Y*=-1;
	}
};
class CStageReboundScreen : public CStage {
public:
	void Reset() {
		New<CPlayerReboundScreen>(0, 0, CVector(0.02f, 0.02f));
	}
};
CStage* StageReboundScreen() { return New<CStageReboundScreen>(); }

//---------------------------------------------------------
class CPlayerReboundBlock : public CPlayer {
public:
	void Move() {
		Position+=Velocity;
		CVector v=Velocity;
		CHitBox hit(this, BlockList, 0.05f);
		if (hit.Left || hit.Right) Velocity.X=-v.X;
		if (hit.Up || hit.Down) Velocity.Y=-v.Y;
	}
};
class CStageReboundBlock : public CStage {
public:
	void Reset() {
		New<CPlayerReboundBlock>(0, 0, CVector(0.02f, 0.01f, 0));
		for (int i=-4; i<=4; i++) {
			New<CBlock>(CVector(1, i*0.2f, 0));
			New<CBlock>(CVector(-1, i*0.2f, 0));
			New<CBlock>(CVector(i*0.2f, 1, 0));
			New<CBlock>(CVector(i*0.2f, -1, 0));
		}
		New<CBlock>(CVector(0.5f, 0, 0));
		New<CBlock>(CVector(-0.5f, 0, 0));
		New<CBlock>(CVector(0, 0.5f, 0));
		New<CBlock>(CVector(0, -0.5f, 0));
	}
};
CStage* StageReboundBlock() { return New<CStageReboundBlock>(); }

//---------------------------------------------------------
class CPlayerPushBlock : public CPlayer {
public:
	void Move() {
		CVector p=Position;
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
		BlockList.Apply([&](CMover* m) {
			CVector mp=m->Position;
			if (CHitBox(m, this, 0.05f)) {
				BlockList.Apply([&](CMover* n) {
					if (m!=n && CHitBox(m, n, 0.05f)) m->Position=mp, Position=p;
				});
			}
		});
	}
};
class CStagePushBlock : public CStage {
public:
	void Reset() {
		New<CPlayerPushBlock>();
		for (int i=0, n=8; i<n; i++) New<CBlock>(AXIS_X*0.6f*RotationZ(float(i)/n));
	}
};
CStage* StagePushBlock() { return New<CStagePushBlock>(); }

//---------------------------------------------------------
class CPlayerPushItem : public CPlayer {
public:
	void Move() {
		CVector p=Position;
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
		ItemList.Apply([&](CMover* m) {
			CVector mp=m->Position, v=m->Position-Position;
			if (Length(v)<0.2f) {
				m->Position=p+Normalize(v)*0.2f;
				ItemList.Apply([&](CMover* n) {
					if (n!=m && Length(n->Position-m->Position)<0.2f) m->Position=mp, Position=p;
				});
			}
		});
	}
};
class CStagePushItem : public CStage {
public:
	void Reset() {
		New<CPlayerPushItem>();
		for (int i=0, n=8; i<n; i++) New<CItem>(AXIS_X*0.6f*RotationZ(float(i)/n));
	}
};
CStage* StagePushItem() { return New<CStagePushItem>(); }

//---------------------------------------------------------
class CCarryItem : public CItem {
public:
	void Move() {
		if (!Owner) {
			Velocity*=0.9f;
			Position+=Velocity;
			PlayerList.Apply([&](CMover* m) {
				if (m->State==0 && Length(m->Position-Position)<0.2f) m->State=1, Owner=m;
			});
		} else {
			Position=Owner->Position+CVector(0, 0.2f);
			if (!Touch[0]) Owner->State=0, Owner=0, Velocity=CVector(0, 0.03f);
		}
	}
};

class CPlayerCarryItem : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
	}
};
class CStageCarryItem : public CStage {
public:
	void Reset() {
		New<CPlayerCarryItem>();
		for (int i=0, n=8; i<n; i++) New<CCarryItem>(AXIS_X*0.6f*RotationZ(float(i)/n));
	}
};
CStage* StageCarryItem() { return New<CStageCarryItem>(); }

//---------------------------------------------------------
class CPlayerUnbreakableObstacle : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
		CHitBox(this, BlockList, 0.05f);
	}
};
class CStageUnbreakableObstacle : public CStage {
public:
	void Reset() {
		New<CPlayerUnbreakableObstacle>();
		for (int i=0; i<10; i++) New<CBlock>(CVector(Random(-1, 1), Random(-1, 1)));
	}
};
CStage* StageUnbreakableObstacle() { return New<CStageUnbreakableObstacle>(); }

//---------------------------------------------------------
class CPlayerBreakableObstacle : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
		BlockList.Apply([&](CMover* m) {
			if (CHitBox(this, m, 0.05f)) m->State=1;
		});
	}
};
class CBlockBreakableObstacle : public CBlock {
public:
	void Move() {
		if (State==1) {
			Position.Y+=Sin(Time*0.1f)*0.01f;
			Color.A-=0.02f;
			if (Color.A<0.1f) Life=0;
			Time++;
		}
	}
};
class CStageBreakableObstacle : public CStage {
public:
	void Reset() {
		New<CPlayerBreakableObstacle>();
		for (int i=0; i<10; i++) New<CBlockBreakableObstacle>(CVector(Random(-1, 1), Random(-1, 1)));
	}
};
CStage* StageBreakableObstacle() { return New<CStageBreakableObstacle>(); }

//---------------------------------------------------------
class CPlayerEnemyPursue : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
	}
};
class CEnemyPursue : public CEnemy {
public:
	void Move() {
		PlayerList.Apply([&](CMover* m) {
			Position.Homing(m->Position, 0.01f);
		});
	}
};
class CStageEnemyPursue : public CStage {
public:
	void Reset() {
		New<CPlayerEnemyPursue>();
		New<CEnemyPursue>(CVector(0, 0.5f, 0));
	}
};
CStage* StageEnemyPursue() { return New<CStageEnemyPursue>(); }

//---------------------------------------------------------
class CPlayerEnemyEscape : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
	}
};
class CEnemyEscape : public CEnemy {
public:
	void Move() {
		PlayerList.Apply([&](CMover* m) {
			Position+=Normalize(Position-m->Position)*0.01f;
		});
		Position.X=max(-0.5f, min(0.5f, Position.X));
		Position.Y=max(-0.75f, min(0.75f, Position.Y));
	}
};
class CStageEnemyEscape : public CStage {
public:
	void Reset() {
		New<CPlayerEnemyEscape>();
		New<CEnemyEscape>(CVector(0, 0.5f, 0));
	}
};
CStage* StageEnemyEscape() { return New<CStageEnemyEscape>(); }

//---------------------------------------------------------
class CPlayerEnemyAttack : public CPlayer {
public:
	void Move() {
		if (Touch[0]) Position.Homing(Touch[0].Position, 0.02f);
	}
};
class CEnemyAttack : public CEnemy {
public:
	void Move() {
		if (State==0) {
			PlayerList.Apply([&](CMover* m) {
				CVector v=m->Position-Position;
				if (Length(v)<0.4f) Velocity=Normalize(v)*0.05f, State=1;
			});
		}
		if (State==1) {
			Scale+=0.002f;
			if (Scale.X>0.12f) State=2;
		}
		if (State==2) {
			Position+=Velocity;
			Velocity*=0.9f;
			if (Length(Velocity)<0.01f) State=3;
		}
		if (State==3) {
			Scale-=0.001f;
			if (Scale.X<=0.1f) Scale=0.1f, State=0;
		}
	}
};
class CStageEnemyAttack : public CStage {
public:
	void Reset() {
		New<CPlayerEnemyAttack>();
		for (int i=0; i<8; i++) New<CEnemyAttack>(CVector(Random(-1, 1), Random(-1, 1)));
	}
};
CStage* StageEnemyAttack() { return New<CStageEnemyAttack>(); }

//---------------------------------------------------------
class CPlayerTilt : public CPlayer {
public:
	void Move() {
		CVector v=Tilt;
		if (abs(v.X)<0.1f) v.X=0;
		if (abs(v.Y)<0.1f) v.Y=0;
		v.X=max(-0.5f, min(0.5f, v.X));
		v.Y=max(-0.5f, min(0.5f, v.Y));
		v.Z=0;
		Position+=v*0.04f;
		Position.X=max(-1.0f, min(1.0f, Position.X));
		Position.Y=max(-1.5f, min(1.5f, Position.Y));
	}
};
class CStageTilt : public CStage {
public:
	void Reset() {
		New<CPlayerTilt>();
	}
};
CStage* StageTilt() { return New<CStageTilt>(); }

