// Copyright (c) 2015 HigPen Works (Kenichiro Matsuura / Yuki Tsukasa). All rights reserved.
#include "Main.h"

//---------------------------------------------------------
class CPlayerSlide : public CPlayer {
public:
	void Move() {
		if (Touch[0]) {
			if (OldTouch[0]) Position+=Touch[0].Position-OldPosition;
			OldPosition=Touch[0].Position;
		}
	}
};
class CStageSlide : public CStage {
public:
	void Reset() {
		New<CPlayerSlide>();
	}
};
CStage* StageSlide() { return New<CStageSlide>(); }

//---------------------------------------------------------
class CPlayerInScreen : public CPlayerSlide {
public:
	void Move() {
		CPlayerSlide::Move();		
		Position.X=max(-0.9f, min(0.9f, Position.X));
		Position.Y=max(-1.4f, min(1.4f, Position.Y));
	}
};
class CStageInScreen : public CStage {
public:
	void Reset() {
		New<CPlayerInScreen>();
	}
};
CStage* StageInScreen() { return New<CStageInScreen>(); }

//---------------------------------------------------------
class CEffectScroll : public CEffect {
public:
	void Move() {
		Scale=0.05f;
		Position.Y-=0.01f;
		if (Position.Y<-1.6f) Life=0;
	}
};
class CStageScroll : public CStage {
public:
	void Move() {
		New<CEffectScroll>(CVector(Random(-1, 1), 1.6f));
	}
};
CStage* StageScroll() { return New<CStageScroll>(); }

//---------------------------------------------------------
class CEffectLayerScroll : public CEffect {
public:
	void Reset() {
		CEffect::Reset();
		Scale=Random(0.03f, 0.05f);
		Velocity.Y=-Scale.Y*0.2f;
	}
	void Move() {
		Position.Y+=Velocity.Y;
		if (Position.Y<-1.6f) Life=0;
	}
};
class CStageLayerScroll : public CStage {
public:
	void Move() {
		New<CEffectLayerScroll>(CVector(Random(-1, 1), 1.6f));
	}
};
CStage* StageLayerScroll() { return New<CStageLayerScroll>(); }

//---------------------------------------------------------
class CBulletBasic : public CBullet {
public:
	void Move() {
		Scale=0.05f;
		Position+=Velocity;
		if (abs(Position.X)>1.1f || abs(Position.Y)>1.6f) Life=0;
	}
};
class CStageBulletBasic : public CStage {
public:
	void Move() {
		New<CBulletBasic>(CVector(0, 0), CVector(0.02f, 0)*RotationZ(Random(0, 1)));
	}
};
CStage* StageBulletBasic() { return New<CStageBulletBasic>(); }

//---------------------------------------------------------
class CPlayerBulletHit : public CPlayerSlide {
public:
	void Move() {
		CPlayerSlide::Move();
		BulletList.Apply([&](CMover* m) {
			if (Length(m->Position-Position)<0.05f) Life=0;
		});
	}
};
class CStageBulletHit : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		New<CBulletBasic>(CVector(0, 0), CVector(0.01f, 0)*RotationZ(Random(0, 1)));
	}
};
CStage* StageBulletHit() { return New<CStageBulletHit>(); }

//---------------------------------------------------------
class CStageBulletAiming : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		if (Time%10==0) {
			PlayerList.Apply([&](CMover* m) {
				CVector from(0, 1.6f);
				New<CBulletBasic>(from, Normalize(m->Position-from)*0.02f);
			});
		}
		Time++;
	}
};
CStage* StageBulletAiming() { return New<CStageBulletAiming>(); }

//---------------------------------------------------------
class CBulletHoming : public CBullet {
public:
	void Move() {
		Scale=0.05f;
		PlayerList.Apply([&](CMover* m) {
			Position.Homing(m->Position, 0.015f);
		});
	}
};
class CStageBulletHoming : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		if (Time%60==0) New<CBulletHoming>(CVector(0, 1.6f));
		Time++;
	}
};
CStage* StageBulletHoming() { return New<CStageBulletHoming>(); }

//---------------------------------------------------------
class CStageBulletNWay : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		if (Time%20==0) NewNWay<CBulletBasic>(CVector(0, 1.6f), CVector(0, -0.02f), 0.2f, 10);
		Time++;
	}
};
CStage* StageBulletNWay() { return New<CStageBulletNWay>(); }

//---------------------------------------------------------
class CStageBulletCircle : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		if (Time%20==0) NewCircle<CBulletBasic>(CVector(0, 0), CVector(0, 0.015f), 20);
		Time++;
	}
};
CStage* StageBulletCircle() { return New<CStageBulletCircle>(); }

//---------------------------------------------------------
class CBulletTurn : public CBullet {
public:
	void Move() {
		Scale=0.05f;
		Position+=Velocity;
		if (Random(0, 1)<0.02f) {
			PlayerList.Apply([&](CMover* m) {
				Velocity=Normalize(m->Position-Position)*0.01f;
			});
		}
	}
};
class CStageBulletTurn : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
		NewCircle<CBulletTurn>(0, CVector(0, 0.01f), 10);
	}
};
CStage* StageBulletTurn() { return New<CStageBulletTurn>(); }

//---------------------------------------------------------
class CBulletSplit : public CBullet {
public:
	void Move() {
		Scale=0.07f;
		Position+=Velocity;
		Time++;
		if (Time>60) {
			NewNWay<CBulletBasic>(Position, Velocity*1.5f, 0.1f, 5);
			Life=0;
		}
	}
};
class CStageBulletSplit : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		if (Time%30==0) New<CBulletSplit>(CVector(0, 1.6f), CVector(0, -0.01f)*RotationZ(Random(-0.1f, 0.1f)));
		Time++;
	}
};
CStage* StageBulletSplit() { return New<CStageBulletSplit>(); }

//---------------------------------------------------------
class CBulletAccelerate : public CBullet {
public:
	void Move() {
		Scale=0.05f;
		Velocity*=1.01f;
		Position+=Velocity;
	}
};
class CStageBulletAccelerate : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		if (Time%10==0) New<CBulletAccelerate>(CVector(0, 1.6f), CVector(0, -0.01f)*RotationZ(Random(-0.1f, 0.1f)));
		Time++;
	}
};
CStage* StageBulletAccelerate() { return New<CStageBulletAccelerate>(); }

//---------------------------------------------------------
class CWeaponBasic : public CWeapon {
public:
	void Move() {
		Scale=0.07f;
		Position+=Velocity;
		if (abs(Position.X)>1.1f || abs(Position.Y)>1.6f) Life=0;
		Rotation=CQuaternion(AXIS_Y, Velocity, AXIS_Z);
	}
};
class CPlayerWeaponBasic : public CPlayerBulletHit {
public:
	void Move() {
		CPlayerBulletHit::Move();
		if (Time%5==0) New<CWeaponBasic>(Position, CVector(0, 0.04f));
		Time++;
	}
};
class CStageWeaponBasic : public CStage {
public:
	void Reset() {
		New<CPlayerWeaponBasic>(CVector(0, -1));
	}
};
CStage* StageWeaponBasic() { return New<CStageWeaponBasic>(); }

//---------------------------------------------------------
class CPlayerWeaponSingle : public CPlayerBulletHit {
public:
	void Move() {
		CPlayerBulletHit::Move();
		if (Touch[0] && !OldTouch[0]) New<CWeaponBasic>(Position, CVector(0, 0.04f));
	}
};
class CStageWeaponSingle : public CStage {
public:
	void Reset() {
		New<CPlayerWeaponSingle>(CVector(0, -1));
	}
};
CStage* StageWeaponSingle() { return New<CStageWeaponSingle>(); }

//---------------------------------------------------------
class CEnemyBasic : public CEnemy {
public:
	void Move() {
		Position+=Velocity;
		WeaponList.Apply([&](CMover* m) {
			if (Length(m->Position-Position)<0.1f) Life--;
		});
		if (Life<=0) Score++;
		if (abs(Position.X)>1.1f || abs(Position.Y)>1.6f) Life=0;
	}
};
class CStageEnemyBasic : public CStage {
public:
	void Reset() {
		UseScore=true;
		New<CPlayerWeaponBasic>(CVector(0, -1));
	}
	void Move() {
		if (Time%30==0) New<CEnemyBasic>(CVector(Random(-1, 1), 1.6f), CVector(0, -0.01f));
		Time++;
	}
};
CStage* StageEnemyBasic() { return New<CStageEnemyBasic>(); }

//---------------------------------------------------------
class CPlayerPullItem : public CPlayerSlide {
public:
	void Move() {
		CPlayerSlide::Move();
		ItemList.Apply([&](CMover* m) {
			float f=Length(m->Position-Position);
			if (f<0.2f) m->Life=0, Score++; else
			if (f<0.6f) m->Position.Homing(Position, 0.03f);
		});
	}
};
class CStagePullItem : public CStage {
public:
	void Reset() {
		UseScore=true;
		New<CPlayerPullItem>();
		for (int i=0; i<20; i++) New<CItem>(CVector(Random(-1, 1), Random(-1.5f, 1.5f)));
	}
};
CStage* StagePullItem() { return New<CStagePullItem>(); }

//---------------------------------------------------------
class CWeaponStraightLaser : public CWeapon {
public:
	void Move() {
		Position.X=Owner->Position.X;
		Position.Y=Owner->Position.Y+(Time+4)*0.04f;
		Time++;
		if (Time>100) Life=0; 
	}
};
class CPlayerStraightLaser : public CPlayerSlide {
public:
	void Move() {
		CPlayerSlide::Move();
		if (Time%3==0) New<CWeaponStraightLaser>(Position)->Owner=this;
		Time++;
	}
};
class CStageStraightLaser : public CStage {
public:
	void Reset() {
		New<CPlayerStraightLaser>();
	}
};
CStage* StageStraightLaser() { return New<CStageStraightLaser>(); }

//---------------------------------------------------------
class CWeaponHomingLaser : public CWeapon {
public:
	void Move() {
		CWeapon::Reset();
		Scale=0.05f;
		if (Type==0) {
			PlayerList.Apply([&](CMover* m) {
				float f=0.01f;
				CVector p=m->Position-Position, v=Velocity;
				CVector a=Normalize(p)*Length(v), b=v*RotationZ(f), c=v*RotationZ(-f);
				if (Dot(v, a)>=Dot(v, b)) Velocity=a; else
				if (Dot(a, b)>=Dot(a, c)) Velocity=b; else Velocity=c;
			});
			Position+=Velocity;
			New<CWeaponHomingLaser>(1, Position, Velocity);
		} else {
			Color.A-=0.02f;
			if (Color.A<=0) Life=0;
		}
		Rotation=CQuaternion(AXIS_Y, Velocity, AXIS_Z);
	}
};
class CStageHomingLaser : public CStage {
public:
	void Reset() {
		New<CPlayerBulletHit>();
	}
	void Move() {
		if (Time%120==0) New<CWeaponHomingLaser>(CVector(0, 1.6f), CVector(0, -0.02f));
		Time++;
	}
};
CStage* StageHomingLaser() { return New<CStageHomingLaser>(); }

//---------------------------------------------------------
class CPlayerLock : public CPlayer {
public:
	void Move() {
		Time++;
		if (Time%10==0) New<CWeaponBasic>(Position, Normalize(Owner->Position-Position)*0.03f);
		if (Touch[0] && !OldTouch[0]) Owner=EnemyList.Next(Owner);
	}
};
class CStageLock : public CStage {
public:
	void Reset() {
		CMover* player=New<CPlayerLock>();
		for (int i=0; i<10; i++) player->Owner=New<CEnemy>(CVector(Random(-1, 1), Random(-1.5f, 1.5f)));
	}
};
CStage* StageLock() { return New<CStageLock>(); }

//---------------------------------------------------------
class CEnemyBomb : public CEnemy {
public:
	void Reset() {
		CEnemy::Reset();
		Life=3;
	}
	void Move() {
		switch (State) {
		case 0:
			if (Random(0, 1)<0.05f) New<CBulletBasic>(Position, CVector(0.02f, 0)*RotationZ(Random(0, 1)));
			break;
		case 1:
			Color.A=0.5f;
			Time++;
			if (Time>100) Color.A=1, State=0;
			break;
		}
	}
};
class CWeaponBomb : public CWeapon {
public:
	void Reset() {
		CWeapon::Reset();
		Texture=GetTexture(L"Bullet.png");
	}
	void Move() {
		switch (State) {
		case 0:
			Velocity*=0.97f;
			Position+=Velocity;
			if (Length(Velocity)<0.001f) State=1;
			break;
		case 1:
			Scale*=1.1f;
			Color.A-=0.02f;
			if (Color.A<=0) Life=0;
			BulletList.Apply([&](CMover* m) {
				if (Length(m->Position-Position)<m->Scale.X+Scale.X) m->Life=0;
			});
			EnemyList.Apply([&](CMover* m) {
				if (m->State==0 && Length(m->Position-Position)<m->Scale.X+Scale.X) m->Life--, m->Time=0, m->State=1;
			});
			break;
		}
	}
};
class CStageBomb : public CStage {
public:
	void Reset() {
		for (int i=0; i<10; i++) New<CEnemyBomb>(CVector(Random(-1, 1), Random(-1.5f, 1.5f)));
	}
	void Move() {
		if (Touch[0] && !OldTouch[0]) New<CWeaponBomb>(CVector(0, -1.6f), CVector(0, 0.05f));
	}
};
CStage* StageBomb() { return New<CStageBomb>(); }

//---------------------------------------------------------
class CEffectBarrier : public CEffect {
public:
	void Reset() {
		CEffect::Reset();
		Scale=0.3f;
	}
	void Move() {
		Position=Owner->Position;
		Rotation*=RotationZ(0.01f);
		BulletList.Apply([&](CMover* m) {
			if (Length(m->Position-Position)<m->Scale.X+Scale.X) m->Life=0, Scale-=0.01f;
		});
		if (Scale.X<0.1f) Life=0;
	}
};
class CStageBarrier : public CStage {
public:
	void Reset() {
		New<CEffectBarrier>()->Owner=New<CPlayerBulletHit>(CVector(0, -1));
	}
	void Move() {
		New<CBulletBasic>(CVector(0, 1.6f), CVector(0, -0.03f)*RotationZ(Random(-0.25f, 0.25f)));
	}
};
CStage* StageBarrier() { return New<CStageBarrier>(); }



