// Copyright (c) 2015 HigPen Works (Kenichiro Matsuura / Yuki Tsukasa). All rights reserved.
#include "Main.h"

//---------------------------------------------------------
class CPieceBasic : public CPlayer {
public:
	void Reset() {
		CPlayer::Reset();
		const wchar_t* texture[]={L"Player.png", L"Enemy.png", L"Bullet.png"};
		Type=(int)RandomLess(0, 3);
		Texture=GetTexture(texture[Type]);
		Scale=0.2f;
		OldPosition=Position;
	}
};

class CPieceSlide : public CPieceBasic {
public:
	void Move() {
		if (State==0) {
			if (!OldTouch[0] && Touch[0] && Length(Touch[0].Position-Position)<Scale.X) State=1;
		}
		if (State==1) {
			Position=Touch[0].Position;
			if (!Touch[0]) State=0;
		}
	}
};
class CStagePieceSlide : public CStage {
public:
	void Reset() {
		New<CPieceSlide>();
	}
};
CStage* StagePieceSlide() { return New<CStagePieceSlide>(); }

//---------------------------------------------------------
class CPieceSwap : public CPieceBasic {
public:
	void Move() {
		if (State==0) {
			if (!OldTouch[0] && Touch[0] && Length(Touch[0].Position-Position)<Scale.X) State=1;
		}
		if (State==1) {
			Color.A=0.5f;
			Position.Homing(Touch[0].Position, Scale.X/2);
			PlayerList.Apply([&](CMover* m) {
				CPlayer* p=(CPlayer*)m;
				if (State==1 && p!=this && Length(p->Position-Position)<Scale.X*1.5f) {
					p->State=State=2;
					swap(p->OldPosition, OldPosition);
				}
			});
			if (!Touch[0]) State=2;
		}
		if (State==2) {
			Color.A=1;
			Position.Homing(OldPosition, Scale.X/5);
			if (OldPosition==Position) State=0;
		}
	}
};
class CStagePieceSwap : public CStage {
public:
	void Reset() {
		for (int x=-2; x<=2; x++) for (int y=-3; y<=3; y++) New<CPieceSwap>(CVector(x*0.4f, y*0.4f));
	}
};
CStage* StagePieceSwap() { return New<CStagePieceSwap>(); }

//---------------------------------------------------------
class CPieceCarry : public CPieceBasic {
public:
	void Move() {
		if (State==0) {
			if (!OldTouch[0] && Touch[0] && Length(Touch[0].Position-Position)<Scale.X) State=1;
		}
		if (State==1) {
			Color.A=0.5f;
			Position.Homing(Touch[0].Position, Scale.X/2);
			PlayerList.Apply([&](CMover* m) {
				CPlayer* p=(CPlayer*)m;
				if (State==1 && p!=this && p->State==0 && Length(p->Position-Position)<Scale.X) {
					p->State=2;
					swap(p->OldPosition, OldPosition);
				}
			});
			if (!Touch[0]) State=2;
		}
		if (State==2) {
			Color.A=1;
			Position.Homing(OldPosition, Scale.X/5);
			if (OldPosition==Position) State=0;
		}
	}
};
class CStagePieceCarry : public CStage {
public:
	void Reset() {
		for (int x=-2; x<=2; x++) for (int y=-3; y<=3; y++) New<CPieceCarry>(CVector(x*0.4f, y*0.4f));
	}
};
CStage* StagePieceCarry() { return New<CStagePieceCarry>(); }

//---------------------------------------------------------
class CPieceChain : public CPieceBasic {
public:
	void Move() {
		if (State==0) {
			Color.A=1;
			if (Touch[0] && Length(Touch[0].Position-Position)<Scale.X) {
				if (!OldTouch[0]) State=1; else {
					PlayerList.Apply([&](CMover* m) {
						if (m!=this && m->Type==Type && m->State>0 && m->State>=State && Length(m->Position-Position)<Scale.X*3) {
							State=max(State, m->State+1); 
						}
					});
				}
			}
		}
		if (State>0) {
			Color.A=0.5f;
			if (!Touch[0]) State=0; else
			if (Length(Touch[0].Position-Position)<Scale.X) {
				PlayerList.Apply([&](CMover* m) {
					if (m!=this && m->State>=State) m->State=0;
				});
			}
		}
	}
};
class CStagePieceChain : public CStage {
public:
	void Reset() {
		for (int x=-2; x<=2; x++) for (int y=-3; y<=3; y++) New<CPieceChain>(CVector(x*0.4f, y*0.4f));
	}
};
CStage* StagePieceChain() { return New<CStagePieceChain>(); }

//---------------------------------------------------------
class CStagePieceChain2 : public CStage {
public:
	void Reset() {
		for (int y=0; y<7; y++) for (int x=0; x<4+y%2; x++) New<CPieceChain>(CVector((x-2)*0.4f+(y+1)%2*0.2f, (y-3)*0.4f));
	}
};
CStage* StagePieceChain2() { return New<CStagePieceChain2>(); }

//---------------------------------------------------------
class CStagePieceLine : public CStage {
public:
	void Reset() {
		for (int x=-2; x<=2; x++) for (int y=-3; y<=3; y++) New<CPieceBasic>(CVector(x*0.4f, y*0.4f));
	}
	void Move() {
		if (!OldTouch[0] && Touch[0]) PlayerList.Kill(), Reset();
		Line();
		PlayerList.Apply([&](CMover* m) { m->Color.A=m->State>0?0.5f:1; });
	}
	void Line() {
		static const int W=5, H=7, N=W*H, L=3;
		CMover* piece[N];
		PlayerList.Apply([&](CMover* m) {
			piece[int(m->Position.X/0.4f+W/2+0.5f)+int(H/2-m->Position.Y/0.4f+0.5f)*W]=m;
		});
		for (int x=0; x<W; x++) for (int y=0; y<H; y++) {
			int i=x+y*W, t=piece[i]->Type, j;
			if (piece[i]->State>0) continue;
			for (j=x+1; j<W && piece[j+y*W]->Type==t; j++) ;
			if (j-x>=L) for (int k=x; k<j; k++) piece[k+y*W]->State=1;
			for (j=y+1; j<H && piece[x+j*W]->Type==t; j++) ;
			if (j-y>=L) for (int k=y; k<j; k++) piece[x+k*W]->State=1;
		}
	}
};
CStage* StagePieceLine() { return New<CStagePieceLine>(); }

//---------------------------------------------------------
class CPieceDrop : public CPieceBasic {
public:
	void Move() {
		if (State==1) {
			Color.A-=0.02f;
			if (Color.A<=0) {
				Life=0;
				New<CPieceDrop>(Position+CVector(0, 2.8f));
				PlayerList.Apply([&](CMover* m) { if (m->State==0) m->State=2; });
			}
		}
		if (State==2) {
			if (Position.Y>1.6f) Position.Y-=0.4f; else Position.Y-=0.04f;
			if (Position.Y<=-1.2f) Position.Y=-1.2f, State=0; else {
				PlayerList.Apply([&](CMover* m) {
					CVector v=Position-m->Position;
					if (m!=this && abs(v.X)<Scale.X && 0<=v.Y && v.Y<=Scale.Y*2) {
						Position.Y=m->Position.Y+Scale.Y*2;
						if (m->State==0) State=0;
					}
				});
			}
		}
	}
};

class CStagePieceDrop : public CStagePieceLine {
public:
	void Reset() {
		for (int x=-2; x<=2; x++) for (int y=-3; y<=3; y++) New<CPieceDrop>(CVector(x*0.4f, y*0.4f));
	}
	void Move() {
		if (!OldTouch[0] && Touch[0]) PlayerList.Kill(), Reset();
		State=0;
		PlayerList.Apply([&](CMover* m) { if (m->State>0) State=1; });
		if (State==0) Line();
	}
};
CStage* StagePieceDrop() { return New<CStagePieceDrop>(); }

//---------------------------------------------------------
class CPieceHarvest : public CPieceBasic {
public:
	void Move() {
		if (State==0) {
			if (Touch[0] && Length(Touch[0].Position-Position)<Scale.X) State=1;
		} else {
			Color.A-=0.02f;
			Position+=CVector(0, 0.01f);
			Scale.X-=0.004f;
			if (Color.A<=0) Life=0;
		}
	}
};

class CStagePieceHarvest : public CStage {
public:
	void Reset() {
		for (int x=-4; x<=4; x++) for (int y=4; y>=-4; y--) New<CPieceHarvest>(CVector(x*0.2f, y*0.3f));
	}
};
CStage* StagePieceHarvest() { return New<CStagePieceHarvest>(); }

//---------------------------------------------------------
class CStagePieceLink : public CStage {
public:
	static const int W=5, H=7, N=W*H, L=4;
	void Reset() {
		for (int x=-2; x<=2; x++) for (int y=-3; y<=3; y++) New<CPieceBasic>(CVector(x*0.4f, y*0.4f));
	}
	void Move() {
		if (!OldTouch[0] && Touch[0]) PlayerList.Kill(), Reset();
		Link();
	}
	void Link() {
		CMover* piece[N];
		PlayerList.Apply([&](CMover* m) {
			piece[int(m->Position.X/0.4f+W/2+0.5f)+int(H/2-m->Position.Y/0.4f+0.5f)*W]=m;
		});
		for (int x=0; x<W; x++) for (int y=0; y<H; y++) {
			if (Count(piece, piece[x+y*W]->Type, x, y)>=L) {
				for (int i=0; i<N; i++) if (piece[i]->State==1) piece[i]->State=2, piece[i]->Color.A=0.5f;
			} else {
				for (int i=0; i<N; i++) piece[i]->State=0;
			}
		}
	}
	int Count(CMover* piece[], int type, int x, int y) {
		int i=x+y*W;
		if (x<0 || x>=W || y<0 || y>=H || piece[i]->Type!=type || piece[i]->State>0) return 0;
		piece[i]->State=1;
		return 1+Count(piece, type, x-1, y)+Count(piece, type, x+1, y)+Count(piece, type, x, y-1)+Count(piece, type, x, y+1);
	}
};
CStage* StagePieceLink() { return New<CStagePieceLink>(); }

//---------------------------------------------------------
class CPieceQuad : public CPlayer {
public:
	void Reset() {
		CPlayer::Reset();
		const wchar_t* texture[]={L"Player.png", L"Enemy.png", L"Weapon.png", L"Bullet.png", L"Item.png", L"Effect.png", L"Block.png"};
		Texture=GetTexture(texture[Type]);
		Scale=0.1f;
	}
};

class CStagePieceQuad : public CStage {
public:
	static const int W=10, H=15, N=W*H, Q=4, TIME=20;
	void Reset() {
		Texture=GetTexture(L"White.png");
		Scale=CVector(0.4f, 1.5f);
		Color=CColor(0, 0, 0, 0.2f);
	}
	void AddPiece(CMover* piece[], CMover* quad[]) {
		static const int xy[]={
			0, 0, 1, 0, 0, -1, 1, -1, 
			0, 0, 0, -1, 0, 1, 0, 2, 
			0, 0, 0, -1, -1, 0, 1, -1,
			0, 0, 0, -1, -1, -1, 1, 0,
			0, 0, -1, 0, 1, 0, 1, -1, 
			0, 0, -1, 0, 1, 0, -1, -1, 
			0, 0, 0, -1, -1, 0, 1, 0, 
		};
		int n=(int)RandomLess(0, 7)*8;
		for (int i=0; i<Q; i++) quad[i]=NewPiece(n/8, W/2+xy[n+i*2], 1+xy[n+i*2+1], i==0?1:2);
		if (Hit(piece, quad, 0, 0, 0)) for (int i=0; i<Q; i++) quad[i]->Life=0;
	}
	CMover* NewPiece(int type, int x, int y, int state) {
		CMover* m=New<CPieceQuad>(type, CVector(x*0.2f-0.9f, 1.4f-y*0.2f));
		m->State=state;
		if (state==3) m->Color.A=0.4f;
		return m;
	}
	int IntX(float x) { return int((x+0.9f)/0.2f+0.5f); }
	int IntY(float y) { return int((1.4f-y)/0.2f+0.5f); }
	int Index(CVector position) { return IntX(position.X)+IntY(position.Y)*W; }
	bool Hit(CMover* piece[], CMover* quad[], int vx, int vy, int rotation) {
		for (int i=0; i<Q; i++) {
			CVector p=quad[i]->Position, c=quad[0]->Position;
			int x=IntX(p.X)+vx, y=IntY(p.Y)+vy;
			int cx=IntX(c.X)+vx, cy=IntY(c.Y)+vy;
			for (int j=0; j<rotation; j++) {
				int t=cx-(y-cy); y=cy+(x-cx); x=t;
			}
			if (x<0 || x>=W || y<0 || y>=H || piece[x+y*W]) return true;
		}
		return false;
	}
	void MakePieceArray(CMover* piece[], CMover* quad[]) {
		for (int i=0; i<N; i++) piece[i]=0;
		int qi=1;
		PlayerList.Apply([&](CMover* m) {
			if (m->State==0) piece[Index(m->Position)]=m;
			if (m->State==1) quad[0]=m;
			if (m->State==2) quad[qi++]=m;
		});
	}
	void RemovePiece(CMover* piece[]) {
		for (int y=0; y<H; y++) for (int x=0; piece[x+y*W]; x++) {
			if (x==W-1) {
				for (int j=0; j<W; j++) {
					int i=j+y*W;
					piece[i]->Life=0;
					for (i-=W; i>=0; i-=W) {
						if (piece[i]) piece[i]->Position.Y-=piece[i]->Scale.Y*2;
						piece[i+W]=piece[i], piece[i]=0;
					}
				}
			}
		}
	}
	void TouchCandidate(CMover* piece[], CMover* quad[]) {
		PlayerList.Apply([&](CMover* m) {
			if (m->State==3) {
				m->Life=0;
				if (quad[0] && !OldTouch[0] && Touch[0] && Length(Touch[0].Position-m->Position)<m->Scale.X) {
					for (int i=0; i<Q; i++) {
						quad[i]->Position+=m->Velocity;
						quad[i]->State=0;
					}
					quad[0]=0;
				}
			}
		});
	}
	void MovePiece(CMover* piece[], CMover* quad[]) {
		if (!OldTouch[0] && Touch[0]) {
			float tx=Touch[0].Position.X;
			if (tx<=-0.4f && !Hit(piece, quad, -1, 0, 0)) for (int i=0; i<Q; i++) quad[i]->Position.X-=quad[i]->Scale.X*2;
			if (tx>=0.4f && !Hit(piece, quad, 1, 0, 0)) for (int i=0; i<Q; i++) quad[i]->Position.X+=quad[i]->Scale.X*2;
			if (abs(tx)<0.4f) {
				int r=1, t=quad[0]->Type;
				if (t==0) r=0; else
				if (t<=3) r=quad[0]->Time%2?3:1;
				if (!Hit(piece, quad, 0, 0, r)) {
					quad[0]->Time++;
					for (int i=1; i<Q; i++) {
						CVector p=quad[i]->Position, c=quad[0]->Position;
						quad[i]->Position=c+(p-c)*RotationZ(r*-0.25f);
					}
				}
			}
		}
	}
	void DropPiece(CMover* piece[], CMover* quad[]) {
		if (Time%TIME==0) {
			if (Hit(piece, quad, 0, 1, 0)) for (int i=0; i<Q; i++) quad[i]->State=0;
			else for (int i=0; i<Q; i++) quad[i]->Position.Y-=quad[i]->Scale.Y*2;
		}
	}
	void AddCandidate(CMover* piece[], CMover* quad[]) {
		if (quad[0]->State>0) {
			int l=W, r=-W;
			for (int i=0; i<Q; i++) {
				int x=IntX(quad[i]->Position.X);
				l=min(l, x), r=max(r, x);
			}
			int w=r-l+2;
			for (int vx=-w*W; vx<=w*W; vx+=w) {
				int vy;
				for (vy=0; !Hit(piece, quad, vx, vy+1, 0); vy++) ;
				if (vy>0) {
					for (int i=0; i<Q; i++) {
						CVector p=quad[i]->Position;
						CMover* m=NewPiece(quad[i]->Type, IntX(p.X)+vx, IntY(p.Y)+vy, 3);
						m->Velocity=CVector((float)vx, (float)-vy)*quad[i]->Scale.X*2;
					}
				}
			}
		}
	}
	void Move() {
		CMover *piece[N], *quad[Q]={0};
		MakePieceArray(piece, quad);
		RemovePiece(piece);
		TouchCandidate(piece, quad);
		if (!quad[0]) {
			AddPiece(piece, quad);
		} else {
			MovePiece(piece, quad);
			DropPiece(piece, quad);
			AddCandidate(piece, quad);
		}
		Time++;
	}
};
CStage* StagePieceQuad() { return New<CStagePieceQuad>(); }

//---------------------------------------------------------
class CPiecePile : public CPieceBasic {
public:
	void Move() {
		if (State==0) {
			CVector op=Position;
			float s=Scale.X*2;
			Velocity.Y-=0.001f;
			Position+=Velocity;
			Velocity*=0.9f;
			PlayerList.Apply([&](CMover* m) {
				CVector v=Position-m->Position;
				if (m!=this && v.Y>=0 && Length(v)<s) Position=m->Position+Normalize(v)*s;
			});
			Position.X=max(-0.8f, min(0.8f, Position.X));
			Position.Y=max(-1.3f, Position.Y);
			Rotation*=RotationZ((op.X-Position.X)*0.9f);
			if (Touch[0] && Length(Touch[0].Position-Position)<Scale.X) State=1;
		} else {
			Color.A-=0.02f;
			if (Color.A<=0) Life=0;
		}
	}
};

class CStagePiecePile : public CStage {
public:
	void Move() {
		for (int i=0; i<2; i++) PlayerList.Move();
		if (Time%15==0) {
			CMover* n=New<CPiecePile>(CVector(Random(-0.8f, 0.8f), 1.7f));
			n->Rotation=CQuaternion(AXIS_Z, Random(0, 1));
			PlayerList.Apply([&](CMover* m) {
				if (n!=m && Length(n->Position-m->Position)<n->Scale.X*2) n->Life=0;
			});
		}
		Time++;
	}
};
CStage* StagePiecePile() { return New<CStagePiecePile>(); }

//---------------------------------------------------------
class CStageStamina : public CStage {
public:
	void Move() {
		int N=10;
		if (Time==60) State=min(N, State+1), Time=0;
		if (!OldTouch[0] && Touch[0]) State=max(0, State-1);
		PlayerList.Kill();
		for (int i=0; i<State; i++) New<CPlayer>(CVector((i-(N-1)/2.0f)*0.2f, 0));
		Time++;
	}
};
CStage* StageStamina() { return New<CStageStamina>(); }

//---------------------------------------------------------
class CStageAddScore : public CStage {
public:
	void Reset() {
		UseScore=true;
		State=Score=Config.GetInt(L"Score");
	}
	void Move() {
		if (!OldTouch[0] && Touch[0]) State+=12345;
		for (int i=1; i<=1000; i*=10) if (State>=Score+i) Score+=i;
		if (Touch[0].Time>60) State=Score=0;
		Config.Set(L"Score", State);
	}
};
CStage* StageAddScore() { return New<CStageAddScore>(); }

